@@ -1,9 +1,9 @@
 #!/usr/bin/perl -w
 
-# Copyright (C) 2002-2014 Jörg Platte <joergplatte@gmx.de>
-# Copyright (C) 2002-2014 Cajus Pollmeier <pollmeier@gonicus.de>
-# Copyright (C) 2013-2014 Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
-# Copyright (C) 2013-2014 Alex Owen <r.alex.owen@gmail.com>
+# Copyright (C) 2002-2015 Jörg Platte <joergplatte@gmx.de>
+# Copyright (C) 2002-2015 Cajus Pollmeier <pollmeier@gonicus.de>
+# Copyright (C) 2013-2015 Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
+# Copyright (C) 2013-2015 Alex Owen <r.alex.owen@gmail.com>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -27,6 +27,8 @@ my $LDAPENABLED = eval "use Net::LDAP; 1" ? '1' : '0';
 
 use Getopt::Std;
 use NetAddr::IP;
+use Data::Validate::IP qw(is_ipv6 is_ipv4);
+use Socket qw(:addrinfo SOCK_RAW AF_INET AF_INET6);
 
 my $SignalCatched=0;
 
@@ -284,42 +286,99 @@ sub simplifyNetworks {
 	my $network;
 	if (@networks) {
 		my $ip;
-		my $onlymacs=1;
+		my $no_macs=1;
 		foreach (@networks) {
-			my $ip;
-			if ($_ =~ /(^[^=]+)=([^=]+)$/ ) {
-				if ($ipv6) {
-					$ip=NetAddr::IP->new6($1) || die "not a valid network: $1\n";
-				} else {
-					$ip=NetAddr::IP->new($1) || die "not a valid network: $1\n";
-				}
-				if (!exists($macs{$ip})) {
-					$macs{$ip}=[];
-				}
-				push (@{$macs{$ip}}, $2);
-			} else {
-				if ($ipv6) {
-					$ip=NetAddr::IP->new6($_) || die "not a valid network: $1\n";
-				} else {
-					$ip=NetAddr::IP->new($_) || die "not a valid network: $1\n";
-				}
-				$onlymacs=0;
-			}
-			push(@netobjects, $ip);
-		};
-		if ($onlymacs==0) {
-			$netref = NetAddr::IP::compactref(\@netobjects);
 
-			@networks=();
-			foreach $network (@$netref) {
-				if (exists($macs{$network})) {
-					foreach $mac (@{$macs{$network}}) {
-						push (@networks, $network."=".$mac);
-					}
+			my $netaddr = '';
+			my $network = '';
+			my $macaddr = '';
+			if ( ($_ =~ /(^[^\/]+)\/([^\/^=]+)$/) || (($ipv6) && is_ipv6($_)) || ((!$ipv6) && is_ipv4($_)) ) {
+
+				if ($ipv6) {
+					$ip=NetAddr::IP->new6($_) || die "not a valid address or network: $_\n";
 				} else {
-					push (@networks, $network);
+					$ip=NetAddr::IP->new($_) || die "not a valid address or network: $_\n";
 				}
-			};
+				push(@netobjects, $ip);
+
+			}
+			elsif ( $_ =~ /(^[^=]+)=([^=]+)$/ ) {
+
+				$netaddr = $1;
+				$macaddr = $2;
+				if ($netaddr =~ /(^[^\/]+)\/([^\/]+)$/) {
+					$network = $1;
+					# FIXME: netmask = $2; TODO: validate netmask
+				} else {
+					$network = $netaddr;
+				}
+
+				if ( (($ipv6) && is_ipv6($network)) || ((!$ipv6) && is_ipv4($network)) ) {
+
+					if ($ipv6) {
+						$ip=NetAddr::IP->new6($netaddr) || die "not a valid address or network: $netaddr\n";
+					} else {
+						$ip=NetAddr::IP->new($netaddr) || die "not a valid address or network: $netaddr\n";
+					}
+					if (!exists($macs{$ip})) {
+						$macs{$ip}=[];
+					}
+					$no_macs = 0;
+					push (@{$macs{$ip}}, $macaddr);
+					push (@netobjects, $ip);
+
+				} else {
+
+					die "Cannot use <dns-name>=<mac-addr> syntax, must be <network-or-ip-addr>=<mac-addr>"
+
+				}
+			}
+			else {
+
+				# resolv "address" that actually is a DNS host name, not an IP address...
+				my $err;
+				my @res;
+
+				if ($ipv6) {
+					( $err, @res ) = getaddrinfo( $_, "", { socktype => SOCK_RAW, family => AF_INET6 } );
+				} else {
+					( $err, @res ) = getaddrinfo( $_, "", { socktype => SOCK_RAW, family => AF_INET } );
+				}
+				die "Cannot getaddrinfo for name '".$_."'- ".$err if $err;
+
+				while ( my $ai = shift @res ) {
+
+					my ( $err, $ipaddr ) = getnameinfo( $ai->{addr}, NI_NUMERICHOST, NIx_NOSERV );
+					die "Cannot getnameinfo - $err" if $err;
+
+					if ($ipv6) {
+						$ip=NetAddr::IP->new6($ipaddr) || die "not a valid address: $ipaddr\n";
+					} else {
+						$ip=NetAddr::IP->new($ipaddr) || die "not a valid address: $ipaddr\n";
+					}
+
+					push (@netobjects, $ip);
+				}
+			}
+
+		}
+
+		# if we don't handle individual MAC addresses on any of the networks, then
+		# we can compact the list of networks
+		if ($no_macs == 1) {
+			@netobjects = NetAddr::IP::Compact(@netobjects);
+		}
+
+		@networks=();
+		foreach $network (@netobjects) {
+			if (exists($macs{$network})) {
+				foreach $mac (@{$macs{$network}}) {
+					push (@networks, $network."=".$mac);
+				}
+			}
+			else {
+				push (@networks, $network);
+			}
 		}
 	}
 	return (@networks);
@@ -563,9 +622,11 @@ sub validateData {
 						$position =~ s/\((.+)\)$//;
 						if (($ipv6) && ($only_proto eq "4")) {
 							print "IPv6 setup: Skipping IPv4-only rule for network \"$position\"\n";
+							$$rule{'Type'} = 'IGNORE-IPV4-ONLY';
 							next;
 						} elsif ((! $ipv6) && ($only_proto eq "6")) {
 							print "IPv4 setup: Skipping IPv6-only rule for network \"$position\"\n";
+							$$rule{'Type'} = 'IGNORE-IPV6-ONLY';
 							next;
 						}
 					}
@@ -585,7 +646,19 @@ sub validateData {
 					}
 				}
 				@simplenetwork = simplifyNetworks (@simplenetwork);
+
+				my $network;
+				for $network (@simplenetwork) {
+					if  ( ( $network =~ /(^[^=]+)=([^=]+)$/ ) && ( ! ( $hashentry =~ /^.*Source$/ ) ) ) {
+						die "MAC address sources don't make sense on destination networks";
+					}
+					if  ( ( $network =~ /(^[^=]+)=([^=]+)$/ ) && ( $$rule{'Type'} eq 'OUTPUT' ) ) {
+						die "MAC address sources don't make sense for outward bound rules";
+					}
+				}
+
 				$$rule{$hashentry} = \@simplenetwork;
+
 			}
 		}
 		foreach (qw(TranslatedSource TranslatedDestination)) {
@@ -960,6 +1033,11 @@ sub genRuleDump {
 	my $chains;
 
 	foreach $rule (@$Rules) {
+
+		if ( ($$rule{'Type'} eq "IGNORE-IPV4-ONLY") || ($$rule{'Type'} eq "IGNORE-IPV6-ONLY") ) {
+			next;
+		}
+
 		my @protocol;
 		my @source;
 		my @destination;
@@ -1130,7 +1208,7 @@ sub genRuleDump {
 			$source=${$$rule{'TranslatedSource'}}[0];
 			$source =~ /([^=]+)/;
 			$source=$1;
-			my $ip = new NetAddr::IP ($source) || die "not a valid network: $source\n";
+			my $ip = NetAddr::IP->new($source) || die "not a valid network: $source\n";
 			my $net=$ip->network();
 			my $bcast = $ip->broadcast();
 			if ($net ne $bcast) {
@@ -1145,7 +1223,7 @@ sub genRuleDump {
 			$destination=${$$rule{'TranslatedDestination'}}[0];
 			$destination =~ /([^=]+)/;
 			$destination=$1;
-			my $ip = new NetAddr::IP ($destination) || die "not a valid network: $destination\n";
+			my $ip = NetAddr::IP->new($destination) || die "not a valid network: $destination\n";
 			my $net=$ip->network();
 			my $bcast = $ip->broadcast();
 			if ($net ne $bcast) {
@@ -1553,7 +1631,7 @@ sub readCommandLine {
 	}
 
 	if ($readldap || $writeldap) {
-		if ($LDAPENABLED == 0) { die "To use LDAP fatures be sure to install Net::LDAP from debain package libnet-ldap-perl" } ;
+		if ($LDAPENABLED == 0) { die "To use LDAP features be sure to install Net::LDAP from the Debian package libnet-ldap-perl" } ;
 		$ldap = Net::LDAP->new($ldapserver) or die "$@";
 		if ($ldapbinddn && ($ldappassword eq "")) {
 			$mesg=$ldap->bind(	$ldapbinddn);
@@ -1901,7 +1979,12 @@ sub writeLdap {
 			$nethost =~ s/\/255.255.255.255//g;
 			if ($nethost =~ /\//) {
 				@entry = ("cn", "$network$counter", "objectClass", "ipNetwork");
-				my $ip = new NetAddr::IP ($nethost) || die "not a valid network: $nethost\n";
+				my $ip;
+				if ($ipv6) {
+					$ip = NetAddr::IP->new6($nethost) || die "not a valid network: $nethost\n";
+				} else {
+					$ip = NetAddr::IP->new($nethost) || die "not a valid network: $nethost\n";
+				}
 				push (@entry, "ipNetworkNumber");
 				push (@entry, $ip->addr());
 				push (@entry, "ipNetmaskNumber");
@@ -1921,9 +2004,17 @@ sub writeLdap {
 				my $mac;
 				if ($nethost =~ /^(.+)=(.+)$/) {
 					$mac=$2;
-					$ip = new NetAddr::IP ($1) || die "not a valid network: $1\n";
+					if ($ipv6) {
+						$ip = NetAddr::IP->new6($1) || die "not a valid network: $1\n";
+					} else {
+						$ip = NetAddr::IP->new($1) || die "not a valid network: $1\n";
+					}
 				} else {
-					$ip = new NetAddr::IP ($nethost) || die "not a valid network: $nethost\n";
+					if ($ipv6) {
+						$ip = NetAddr::IP->new6($nethost) || die "not a valid network: $nethost\n";
+					} else {
+						$ip = NetAddr::IP->new($nethost) || die "not a valid network: $nethost\n";
+					}
 				}
 				push (@entry, "ipHostNumber");
 				push (@entry, $ip->addr());
